@@ -13,6 +13,7 @@ package org.eclipse.cdt.managedbuilder.internal.core;
 import java.net.MalformedURLException;
 import java.net.URL;
 import java.util.ArrayList;
+import java.util.Collection;
 import java.util.HashMap;
 import java.util.Iterator;
 import java.util.List;
@@ -26,12 +27,12 @@ import org.eclipse.cdt.managedbuilder.core.IBuildObject;
 import org.eclipse.cdt.managedbuilder.core.IConfiguration;
 import org.eclipse.cdt.managedbuilder.core.IEnvVarBuildPath;
 import org.eclipse.cdt.managedbuilder.core.IHoldsOptions;
-import org.eclipse.cdt.managedbuilder.core.IManagedProject;
-import org.eclipse.cdt.managedbuilder.core.IOptionApplicability;
 import org.eclipse.cdt.managedbuilder.core.IInputType;
 import org.eclipse.cdt.managedbuilder.core.IManagedCommandLineGenerator;
 import org.eclipse.cdt.managedbuilder.core.IManagedConfigElement;
+import org.eclipse.cdt.managedbuilder.core.IManagedProject;
 import org.eclipse.cdt.managedbuilder.core.IOption;
+import org.eclipse.cdt.managedbuilder.core.IOptionApplicability;
 import org.eclipse.cdt.managedbuilder.core.IOptionCategory;
 import org.eclipse.cdt.managedbuilder.core.IOutputType;
 import org.eclipse.cdt.managedbuilder.core.IProjectType;
@@ -46,14 +47,15 @@ import org.eclipse.cdt.managedbuilder.internal.macros.MacroResolver;
 import org.eclipse.cdt.managedbuilder.macros.BuildMacroException;
 import org.eclipse.cdt.managedbuilder.macros.IBuildMacroProvider;
 import org.eclipse.cdt.managedbuilder.makegen.IManagedDependencyGenerator;
+import org.eclipse.cdt.managedbuilder.makegen.IManagedDependencyGeneratorType;
 import org.eclipse.core.resources.IProject;
 import org.eclipse.core.resources.ProjectScope;
-import org.eclipse.core.runtime.IConfigurationElement;
 import org.eclipse.core.runtime.CoreException;
+import org.eclipse.core.runtime.IConfigurationElement;
 import org.eclipse.core.runtime.IExtension;
 import org.eclipse.core.runtime.IExtensionPoint;
-import org.eclipse.core.runtime.Path;
 import org.eclipse.core.runtime.IPath;
+import org.eclipse.core.runtime.Path;
 import org.eclipse.core.runtime.Platform;
 import org.eclipse.core.runtime.PluginVersionIdentifier;
 import org.eclipse.core.runtime.content.IContentType;
@@ -74,6 +76,9 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	public static final String DEFAULT_PATTERN = "${COMMAND} ${FLAGS} ${OUTPUT_FLAG}${OUTPUT_PREFIX}${OUTPUT} ${INPUTS}"; //$NON-NLS-1$
 	public static final String DEFAULT_CBS_PATTERN = "${COMMAND}"; //$NON-NLS-1$
 
+	//property name for holding the rebuild state
+	private static final String REBUILD_STATE = "rebuildState";  //$NON-NLS-1$
+
 	private static final String DEFAULT_SEPARATOR = ","; //$NON-NLS-1$
 	//private static final IOptionCategory[] EMPTY_CATEGORIES = new IOptionCategory[0];
 	//private static final IOption[] EMPTY_OPTIONS = new IOption[0];
@@ -86,7 +91,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	
 	//  Superclass
 	//  Note that superClass itself is defined in the base and that the methods
-	//  getSuperClass() and setSuperClass(), defined in Tool must be used to 
+	//  getSuperClass() and setSuperClassInternal(), defined in Tool must be used to 
 	//  access it. This avoids widespread casts from IHoldsOptions to ITool.
 	private String superClassId;
 	//  Parent and children
@@ -116,7 +121,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	private IConfigurationElement commandLineGeneratorElement = null;
 	private IManagedCommandLineGenerator commandLineGenerator = null;
 	private IConfigurationElement dependencyGeneratorElement = null;
-	private IManagedDependencyGenerator dependencyGenerator = null;
+	private IManagedDependencyGeneratorType dependencyGenerator = null;
 	private URL iconPathURL;	
 	//  Miscellaneous
 	private boolean isExtensionTool = false;
@@ -124,6 +129,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	private boolean resolved = resolvedDefault;
 	private IConfigurationElement previousMbsVersionConversionElement = null;
 	private IConfigurationElement currentMbsVersionConversionElement = null;
+	private boolean rebuildState;
 
 	/*
 	 *  C O N S T R U C T O R S
@@ -199,7 +205,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	public Tool(ToolChain parent, ITool superClass, String Id, String name, boolean isExtensionElement) {
 		super(resolvedDefault);
 		this.parent = parent;
-		setSuperClass(superClass);
+		setSuperClassInternal(superClass);
 		setManagedBuildRevision(parent.getManagedBuildRevision());
 		if (getSuperClass() != null) {
 			superClassId = getSuperClass().getId();
@@ -215,6 +221,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 			ManagedBuildManager.addExtensionTool(this);
 		} else {
 			setDirty(true);
+			setRebuildState(true);
 		}
 	}
 
@@ -232,7 +239,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	public Tool(ResourceConfiguration parent, ITool superClass, String Id, String name, boolean isExtensionElement) {
 		super(resolvedDefault);
 		this.parent = parent;
-		setSuperClass( superClass );
+		setSuperClassInternal( superClass );
 		setManagedBuildRevision(parent.getManagedBuildRevision());
 		if (getSuperClass() != null) {
 			superClassId = getSuperClass().getId();
@@ -247,6 +254,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 			ManagedBuildManager.addExtensionTool(this);
 		} else {
 			setDirty(true);
+			setRebuildState(true);
 		}
 	}
 
@@ -287,6 +295,11 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 				addOutputType(outputType);
 			}
 		}
+		
+		String rebuild = PropertyManager.getInstance().getProperty(this, REBUILD_STATE);
+		if(rebuild == null || Boolean.valueOf(rebuild).booleanValue())
+			rebuildState = true;
+
 	}
 
 	/**
@@ -300,9 +313,9 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		super(resolvedDefault);
 		this.parent = parent;
 		if (toolSuperClass != null) {
-			setSuperClass( toolSuperClass );
+			setSuperClassInternal( toolSuperClass );
 		} else {
-			setSuperClass( tool.getSuperClass() );
+			setSuperClassInternal( tool.getSuperClass() );
 		}
 		if (getSuperClass() != null) {
 			superClassId = getSuperClass().getId();
@@ -420,6 +433,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		}	
 		
         setDirty(true);
+		setRebuildState(true);
 	}
 	
 	/*
@@ -571,9 +585,9 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		if (superClassId != null && superClassId.length() > 0) {
 			if( getParent() instanceof IResourceConfiguration ) {
 				IResourceConfiguration resConfig = (IResourceConfiguration) getParent();
-				setSuperClass( resConfig.getParent().getTool(superClassId) );
+				setSuperClassInternal( resConfig.getParent().getTool(superClassId) );
 			} else {
-				setSuperClass( ManagedBuildManager.getExtensionTool(superClassId) );			
+				setSuperClassInternal( ManagedBuildManager.getExtensionTool(superClassId) );			
 			}
 			
 			// Check for migration support
@@ -867,7 +881,9 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 				// Save as URL in string form
 				element.setAttribute(IOptionCategory.ICON, iconPathURL.toString());
 			}
-			
+
+			saveRebuildState();
+
 			// I am clean now
 			isDirty = false;
 		} catch (Exception e) {
@@ -1077,13 +1093,21 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		}
 		return iconPathURL;
 	}
+	
+	/* (non-Javadoc)
+	 * @see org.eclipse.cdt.core.build.managed.IOptionCategory#getOptions()
+	 */
+	public Object[][] getOptions(IConfiguration configuration, IHoldsOptions optionHolder) {
+		if (optionHolder != this) return null;
+		return getOptions(configuration);
+	}
 
 	/* (non-Javadoc)
-	 * @see org.eclipse.cdt.core.build.managed.IOptionCategory#getOptions(org.eclipse.cdt.core.build.managed.ITool)
+	 * @see org.eclipse.cdt.core.build.managed.IOptionCategory#getOptions()
 	 */
 	public Object[][] getOptions(IConfiguration configuration) {
 		// Find the child of the configuration that represents the same tool.
-		// It could the tool itself, or a "sub-class" of the tool.
+		// It could be the tool itself, or a "sub-class" of the tool.
 		if (configuration != null) {
 			ITool[] tools = configuration.getTools();
 			return getOptions(tools);
@@ -1092,6 +1116,17 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		}
 	}
 	
+	/* (non-Javadoc)
+	 * @see org.eclipse.cdt.core.build.managed.IOptionCategory#getOptions()
+	 */
+	public Object[][] getOptions(IResourceConfiguration resConfig, IHoldsOptions optionHolder) {
+		if (optionHolder != this) return null;
+		return getOptions(resConfig);
+	}
+	
+	/* (non-Javadoc)
+	 * @see org.eclipse.cdt.core.build.managed.IOptionCategory#getOptions()
+	 */
 	public Object[][] getOptions(IResourceConfiguration resConfig) {
 		ITool[] tools = resConfig.getTools();
 		return getOptions(tools);
@@ -1232,9 +1267,23 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	 * Access function to set the superclass element that is defined in
 	 * the base class.
 	 */
-	private void setSuperClass(ITool superClass) {
+	private void setSuperClassInternal(ITool superClass) {
 		this.superClass = superClass;
 	}
+	
+	public void setSuperClass(ITool superClass) {
+		if ( this.superClass != superClass ) {
+			this.superClass = superClass;
+			if ( this.superClass == null) {
+				superClassId = null;
+			} else {
+				superClassId = this.superClass.getId();
+			}
+		
+			if(!isExtensionElement())
+				setDirty(true);
+		}			
+	}
 
 	/* (non-Javadoc)
 	 * @see org.eclipse.cdt.core.build.managed.ITool#getName()
@@ -1783,14 +1832,22 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	 */
 	public IManagedDependencyGenerator getDependencyGenerator() {
 		if (dependencyGenerator != null) {
-			return dependencyGenerator;
+			if (dependencyGenerator instanceof IManagedDependencyGenerator)
+				return (IManagedDependencyGenerator)dependencyGenerator;
+			else
+				return null;
 		}
 		IConfigurationElement element = getDependencyGeneratorElement();
 		if (element != null) {
 			try {
 				if (element.getAttribute(DEP_CALC_ID) != null) {
-					dependencyGenerator = (IManagedDependencyGenerator) element.createExecutableExtension(DEP_CALC_ID);
-					return dependencyGenerator;
+					dependencyGenerator = (IManagedDependencyGeneratorType) element.createExecutableExtension(DEP_CALC_ID);
+					if (dependencyGenerator != null) {
+						if (dependencyGenerator instanceof IManagedDependencyGenerator)
+							return (IManagedDependencyGenerator)dependencyGenerator;
+						else
+							return null;
+					}
 				}
 			} catch (CoreException e) {}
 		}
@@ -1800,7 +1857,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 	/* (non-Javadoc)
 	 * @see org.eclipse.cdt.managedbuilder.core.ITool#getDependencyGeneratorForExtension()
 	 */
-	public IManagedDependencyGenerator getDependencyGeneratorForExtension(String sourceExt) {
+	public IManagedDependencyGeneratorType getDependencyGeneratorForExtension(String sourceExt) {
 		if (dependencyGenerator != null) {
 			return dependencyGenerator;
 		}
@@ -1808,7 +1865,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		if (element != null) {
 			try {
 				if (element.getAttribute(DEP_CALC_ID) != null) {
-					dependencyGenerator = (IManagedDependencyGenerator) element.createExecutableExtension(DEP_CALC_ID);
+					dependencyGenerator = (IManagedDependencyGeneratorType) element.createExecutableExtension(DEP_CALC_ID);
 					return dependencyGenerator;
 				}
 			} catch (CoreException e) {}
@@ -1962,6 +2019,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		if (cmd == null || command == null || !cmd.equals(command)) {
 			command = cmd;
 			isDirty = true;
+			setRebuildState(true);
 			return true;
 		} else {
 			return false;
@@ -1975,6 +2033,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		if (pattern == null && commandLinePattern == null) return;
 		if (pattern == null || commandLinePattern == null || !pattern.equals(commandLinePattern)) {
 			commandLinePattern = pattern;
+			setRebuildState(true);
 			isDirty = true;
 		}
 	}
@@ -1986,6 +2045,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		if (flag == null && outputFlag == null) return;
 		if (outputFlag == null || flag == null || !(flag.equals(outputFlag))) {
 			outputFlag = flag;
+			setRebuildState(true);
 			isDirty = true;
 		}
 	}
@@ -1997,6 +2057,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		if (prefix == null && outputPrefix == null) return;
 		if (outputPrefix == null || prefix == null || !(prefix.equals(outputPrefix))) {
 			outputPrefix = prefix;
+			setRebuildState(true);
 			isDirty = true;
 		}
 	}
@@ -2362,7 +2423,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 			resolved = true;
 			// Resolve superClass
 			if (superClassId != null && superClassId.length() > 0) {
-				setSuperClass( ManagedBuildManager.getExtensionTool(superClassId) );
+				setSuperClassInternal( ManagedBuildManager.getExtensionTool(superClassId) );
 				if (getSuperClass() == null) {
 					// Report error
 					ManagedBuildManager.OutputResolveError(
@@ -2370,6 +2431,10 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 							superClassId,
 							"tool",	//$NON-NLS-1$
 							getId());
+				} else {
+					//  All of our superclasses must be resolved in order to properly
+					//  resolve options to option categories   
+					((Tool)getSuperClass()).resolveReferences();
 				}
 			}
 			//  Resolve HoldsOptions 
@@ -2567,14 +2632,15 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 
 			// for each element in the 'subMap',
 			// check the 'versionsSupported' attribute whether the given
-			// builder version is supported
+			// tool version is supported
 
 			String baseId = ManagedBuildManager
 					.getIdFromIdAndVersion(superClassId);
 			String version = ManagedBuildManager
 					.getVersionFromIdAndVersion(superClassId);
 
-			ITool[] toolElements = (ITool[]) subMap.values().toArray();
+			Collection c = subMap.values();
+			ITool[] toolElements = (ITool[]) c.toArray(new ITool[c.size()]);
 
 			for (int i = 0; i < toolElements.length; i++) {
 				ITool toolElement = toolElements[i];
@@ -2613,7 +2679,7 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 								// If control comes here means that superClass
 								// is null.
 								// So, set the superClass to this tool element
-								setSuperClass(toolElement);
+								setSuperClassInternal(toolElement);
 								superClassId = getSuperClass().getId();
 								isExists = true;
 								break;
@@ -2780,4 +2846,64 @@ public class Tool extends HoldsOptions implements ITool, IOptionCategory {
 		return globalSpecs;		
 	}
 
+	/* (non-Javadoc)
+	 * @see org.eclipse.cdt.managedbuilder.internal.core.HoldsOptions#needsRebuild()
+	 */
+	public boolean needsRebuild() {
+		if(rebuildState)
+			return true;
+		
+		// Check my children
+		List typeElements = getInputTypeList();
+		Iterator iter = typeElements.listIterator();
+		while (iter.hasNext()) {
+			InputType type = (InputType) iter.next();
+			if (type.needsRebuild()) return true;
+		}
+		typeElements = getOutputTypeList();
+		iter = typeElements.listIterator();
+		while (iter.hasNext()) {
+			OutputType type = (OutputType) iter.next();
+			if (type.needsRebuild()) return true;
+		}
+
+		return super.needsRebuild();
+	}
+
+	/* (non-Javadoc)
+	 * @see org.eclipse.cdt.managedbuilder.internal.core.HoldsOptions#setRebuildState(boolean)
+	 */
+	public void setRebuildState(boolean rebuild) {
+		if(isExtensionElement() && rebuild)
+			return;
+		
+		if(rebuildState != rebuild){
+			rebuildState = rebuild;
+			saveRebuildState();
+		}
+		
+		if(!rebuild){
+			super.setRebuildState(rebuild);
+			
+			if (!rebuild) {
+				List typeElements = getInputTypeList();
+				Iterator iter = typeElements.listIterator();
+				while (iter.hasNext()) {
+					InputType type = (InputType) iter.next();
+					type.setRebuildState(false);
+				}
+				typeElements = getOutputTypeList();
+				iter = typeElements.listIterator();
+				while (iter.hasNext()) {
+					OutputType type = (OutputType) iter.next();
+					type.setRebuildState(false);
+				}
+			}
+		}
+	}
+	
+	private void saveRebuildState(){
+		PropertyManager.getInstance().setProperty(this, REBUILD_STATE, Boolean.toString(needsRebuild()));
+	}
+
 }
